
from typing import List
from azure.storage.blob import BlobServiceClient
from . import storage_backend

class AzureBlobStorageBackend(storage_backend.StorageBackend[str, bytes]):
    def __init__(self, connection_string: str, container_name: str):
        self._container_client = (BlobServiceClient
                .from_connection_string(connection_string)
                .get_container_client(container_name))

    def _blob_client(self, name: str):
        return self._container_client.get_blob_client(name)

    def store(self, key: str, value: bytes) -> None:
        """
        retrieve
        ========

        parameters:
            key (str)
            value (bytes)

        Stores the data in a blob, overwriting it if it exists.
        """

        if (client := self._blob_client(key)).exists():
            client.delete_blob()

        client.upload_blob(value)

    def retrieve(self, key: str) -> bytes:
        """
        retrieve
        ========

        parameters:
            key (str)
        returns:
            bytes

        Fetches the contents of a blob
        """
        if (client := self._blob_client(key)).exists():
            return client.download_blob().readall()
        else:
            raise KeyError(f"{key} does not exist")

    def exists(self, key: str) -> bool:
        """
        exists
        ===========

        parameters:
            path (str): The path to a file on the server

        returns:
            bool: Whether file exists or not

        For a given key check if the blob exists
        """
        return self._blob_client(key).exists()

    def keys(self) -> List[str]:
        """
        keys
        ====

        returns:
            List[str]

        Returns the names of all of the blobs currently stored.
        """
        return [blob.name for blob in self._container_client.list_blobs()]

    def __del__(self):
        self._container_client.close()




from typing import Union, Dict, List, Optional
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from . import storage_backend

KeyType = Union[str, int]
JsonSerializable = Union[
        Dict["JsonSerializable","JsonSerializable"],
        List["JsonSerializable"],
        str, int, float, bool, None]

class Sql(storage_backend.StorageBackend[KeyType, JsonSerializable]):

    def __init__(self, engine: Engine, table_name: str, schema: Optional[str] = None):
        self._engine = engine
        md_args = {"schema": schema} if schema is not None else {}
        self._metadata = sa.MetaData(**md_args)
        self._table = sa.Table(table_name, self._metadata, autoload_with = self._engine)
        self._assert_one_pk()

    @property
    def _primary_key(self):
        return self._table.primary_key.columns[0]

    @property
    def fields(self):
        return [(c.name, c.type) for c in self._table.columns]

    def store(self, key: KeyType, value: JsonSerializable) -> None:
        with self._engine.connect() as con:
            values = {self._primary_key.name: key, **value}
            query = self._table.insert().values(**values)
            con.execute(query)

    def retrieve(self, key: KeyType) -> JsonSerializable:
        with self._engine.connect() as con:
            query = self._table.select().where(self._primary_key == key)
            res = con.execute(query).fetchone()
            if res is not None:
                data = dict(res)
                del(data[self._primary_key.name])
                return data
            else:
                return None

    def exists(self, key: KeyType) -> bool:
        return self.retrieve(key) is not None

    def keys(self):
        with self._engine.connect() as con:
            return con.execute(sa.select(self._primary_key)).fetchall()

    def _assert_one_pk(self):
        try:
            _,*excess = [c.name for c in self._table.primary_key.columns]
            assert len(excess) == 0
        except AssertionError:
            raise ValueError("The database table has a composite primary key, which is not currently supported")



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")
U = TypeVar("U")

class StorageBackend(ABC, Generic[T, U]):
    @abstractmethod
    def store(self, key: T, value: U) -> None:
        raise NotImplementedError()

    @abstractmethod
    def retrieve(self, key: T) -> U:
        raise NotImplementedError()

    @abstractmethod
    def exists(self, key: T) -> bool:
        """
        exists
        ===========

        parameters:
            path (str): Uhe path to a file on the server

        returns:
            bool: Whether file exists or not

        For a given path check if file exists
        """

        raise NotImplementedError()

    @abstractmethod
    def keys(self):
        raise NotImplementedError()
